import enum
import re
from timeit import default_timer

import numpy as np
import xarray as xr
from calngUtils import device as device_utils, trackers
from geometryDevices import utils as geom_utils
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    INT32_ELEMENT,
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    TABLE_ELEMENT,
    ChannelMetaData,
    Epochstamp,
    Hash,
    ImageData,
    MetricPrefix,
    Schema,
    Timestamp,
    Trainstamp,
    Unit,
)
from TrainMatcher import TrainMatcher

from . import preview_utils, scenes, schemas
from ._version import version as deviceVersion


class BridgeOutputOptions(enum.Enum):
    MATCHED = "matched"
    ASSEMBLED = "assembled"


def module_index_schema():
    schema = Schema()
    (
        BOOL_ELEMENT(schema)
        .key("select")
        .assignmentOptional()
        .defaultValue(True)
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("source")
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("regexUsed")
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),

        INT32_ELEMENT(schema)
        .key("index")
        .assignmentOptional()
        .defaultValue(-1)
        .reconfigurable()
        .commit(),
    )
    return schema


@device_utils.with_unsafe_get
@KARABO_CLASSINFO("DetectorAssembler", deviceVersion)
class DetectorAssembler(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            .setNewDefaultValue(["overview", "trainMatcherScene"])
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("timeOfFlight")
            .displayedName("Time of flight")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .warnHigh(1000)
            .info("Time of flight exceeding 1 s")
            .needsAcknowledging(False)
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("processingTime")
            .displayedName("Processing time")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .commit(),

            STRING_ELEMENT(expected)
            .key("imageDataPath")
            .assignmentOptional()
            .defaultValue("image.data")
            .commit(),

            STRING_ELEMENT(expected)
            .key("imageMaskPath")
            .assignmentOptional()
            .defaultValue("image.mask")
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("assembledOutput")
            .dataSchema(schemas.preview_schema(wrap_image_in_imagedata=False))
            .commit(),

            STRING_ELEMENT(expected)
            .key("outputForBridgeOutput")
            .displayedName("Bridge output")
            .description(
                "The bridge can send the same data as any one of the three Karabo "
                "output channels. If this setting is 'matched', it will follow "
                "'output', if 'assembled', it will follow 'assembledOutput' and if "
                "'preview', it will follow 'preview.output'."
            )
            .options(",".join(option.value for option in BridgeOutputOptions))
            .assignmentOptional()
            .defaultValue("matched")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("geometryDevice")
            .displayedName("Geometry device")
            .description(
                "The the device which will provide geometry. The device is expected to "
                "have current geometry as a string (for now using custom non-robust "
                "serialization scheme) element named serializedGeometry."
            )
            .assignmentMandatory()
            .commit(),

            STRING_ELEMENT(expected)
            .key("sourceIndexPattern")
            .displayedName("Source index regex")
            .description(
                "Regular expression used to extract the module index (used with "
                "geometry for assembly) from source names. This must contain a capture "
                "group named 'index'; an example expression could be "
                "'.*\\/MY_DETECTOR\\/(?P<index>\\d+).*' "
                "(note the '.*' in the prefix; the expression must match from the "
                "start of the source name). If no source index pattern is provided, "
                "some default patterns for correction devices, XTDF detectors, or DAQ "
                "outputs (as typically used with JUNGFRAU detectors) will be tried."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            TABLE_ELEMENT(expected)
            .key("sourceToModuleIndex")
            .displayedName("Module index mapping")
            .description(
                "This table shows how the sourceIndexPattern (or if not set, one of "
                "the default patterns) extracted module indices for the configured "
                "(and selected) sources. Empty indices indicate sources which did not "
                "match the regular expression and will not be used in assembly."
            )
            .setColumns(module_index_schema())
            .assignmentOptional()
            .defaultValue([])
            .reconfigurable()
            .commit(),
        )
        preview_utils.PreviewFriend.add_schema(expected, "preview", create_node=True)
        (
            OVERWRITE_ELEMENT(expected)
            .key("preview.flipSS")
            .setNewDefaultValue(True)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.flipFS")
            .setNewDefaultValue(True)
            .commit(),
        )

    def __init__(self, conf):
        super().__init__(conf)
        self.info.merge(Hash("timeOfFlight", 0, "processingTime", 0))
        self._tof_tracker = trackers.ExponentialMovingAverage(alpha=0.3)
        self._processing_time_tracker = trackers.ExponentialMovingAverage(
            alpha=0.3
        )
        self.registerSlot(self.slotReceiveGeometry)
        self._need_to_update_source_index_mapping = True

    def initialization(self):
        super().initialization()

        self._preview_friend = preview_utils.PreviewFriend(self)
        self._image_data_path = self.get("imageDataPath")
        self._image_mask_path = self.get("imageMaskPath")
        self._geometry = None
        # set up source to index mapping
        self._merge_source_to_index_from_regex()
        self._set_source_to_index_from_table()

        self.KARABO_SLOT(self.requestScene)

        self.signalSlotable.connect(
            self.get("geometryDevice"),
            "signalNewGeometry",
            "",  # slot device ID (default: self)
            "slotReceiveGeometry",
        )

        self.assembled_output = self.signalSlotable.getOutputChannel("assembledOutput")
        self.start()

    def _source_to_index_from_regex(self):
        """Helper function to run through either user supplied or built-in regex and
        extract mapping of source -> (matching regex, module index) from current table
        of sources. If relying on built-in regexes, will use the first one to match
        anything on all sources."""
        res = {}
        sources = [
            row["source"].partition("@")[0]
            for row in self.get("sources")
            if row["select"]
        ]
        if (source_pattern := self.get("sourceIndexPattern")):
            source_re = re.compile(source_pattern)
            res = {
                source: (source_pattern, int(match.group("index")))
                for source in sources
                if (match := source_re.match(source)) is not None
            }
        else:
            for (source_pattern, extractor) in [
                (
                    r".*\/DET\/(\d+)CH0:xtdf",  # XTDF-detector source
                    lambda match: int(match.group(1)),
                ),
                (
                    r".*\/CORRECT(\d+)_Q(\d)M(\d)",  # typical QxMy for big detectors
                    lambda match: (int(match.group(2)) - 1) * 4
                    + (int(match.group(3)) - 1),
                ),
                (
                    r".*\/DET\/.*?(\d+):daqOutput",  # typical JF receiver source
                    lambda match: int(match.group(1)),
                    # don't subtract 1; JUNGFRAUGeometry does this for us
                ),
                (
                    r".*\/CORRECT(\d+)_M(\d+)",  # typical JF correction device
                    lambda match: int(match.group(2)),
                    # don't subtract 1; JUNGFRAUGeometry does this for us
                ),
                (
                    r".*\/CORRECT(\d+)",  # generic correction device enumeration
                    lambda match: int(match.group(1)),
                ),
            ]:
                source_re = re.compile(source_pattern)
                # warning: will accept partial match, careful about mixing sources
                matches = [source_re.match(source) for source in sources]
                if any(match is not None for match in matches):
                    res = {
                        source: (source_pattern, extractor(match))
                        for (source, match) in zip(sources, matches)
                        if match is not None
                    }
                    break
        for source in sources:
            if source not in res:
                res[source] = (None, None)
        return res

    def _merge_source_to_index_from_regex(self):
        """Helper function to fill out the source module index table. It will use
        _source_to_index_from_regex and add to the sourceToModuleIndex table any sources
        not yet set (ex. new source added to sources / device just initialized / some
        sources manually set)."""
        table = self.get("sourceToModuleIndex")
        known_sources = {row["source"] for row in table if row["select"]}
        used_indices = {row["index"] for row in table if row["select"]}
        for source, (
            source_pattern,
            index,
        ) in self._source_to_index_from_regex().items():
            if source in known_sources:
                # already in table - maybe manually set
                continue
            elif index is None:
                # was not matched by chosen regex (if any)
                table.append(Hash("select", False, "source", source, "index", -1))
            elif index in used_indices:
                # was matched, but module index used by other source
                self.log.WARN(f"Index {index} already used, deselecting {source}")
                table.append(
                    Hash(
                        "select",
                        False,
                        "source",
                        source,
                        "regexUsed",
                        source_pattern,
                        "index",
                        index,
                    )
                )
            else:
                table.append(
                    Hash(
                        "select",
                        True,
                        "source",
                        source,
                        "regexUsed",
                        source_pattern,
                        "index",
                        index,
                    )
                )
                used_indices.add(index)
        self.set(
            "sourceToModuleIndex",
            table
        )

    def _set_source_to_index_from_table(self):
        """Sets internal source to module index mapping based on the values currently in
        the sourceToModuleIndex table. Should be called whenever the table is updated.
        """
        mapping = {
            row["source"]: row["index"]
            for row in self.get("sourceToModuleIndex")
            if row["select"]
        }
        assert all(value >= 0 for value in mapping.values())
        assert len(mapping) == len(set(mapping.values()))
        self._source_to_index = mapping
        self._need_to_update_source_index_mapping = False

    def requestScene(self, params):
        scene_name = params.get("name", default="")
        if scene_name == "trainMatcherScene":
            params["name"] = "scene"
            return super().requestScene(params)
        if scene_name != "overview":
            self.log.INFO(
                f"Unexpected scene name {scene_name} requested, "
                "will just return overview"
            )
        payload = Hash("name", scene_name, "success", True)
        payload["data"] = scenes.detector_assembler_overview(
            device_id=self.getInstanceId(),
            schema=self.getFullSchema(),
        )
        self.reply(
            Hash(
                "type",
                "deviceScene",
                "origin",
                self.getInstanceId(),
                "payload",
                payload,
            )
        )

    def slotReceiveGeometry(self, device_id, serialized_geometry):
        self.log.INFO(f"Received geometry from {device_id}")
        try:
            self._geometry = geom_utils.deserialize_geometry(serialized_geometry)
        except Exception as e:
            self.log.WARN(f"Failed to deserialize geometry; {e}")
            return
        # TODO: test with multiple memory cells (extra geom notion of extra dimensions)
        # NOTE: could almost test n_modules against _source_to_index JF inconsistent

    def on_matched_data(self, train_id, sources):
        ts_start = default_timer()
        if self._geometry is None:
            self.log.WARN("Have not received a geometry yet, will not send anything")
            return
        my_timestamp = Timestamp(Epochstamp(), Trainstamp(train_id))
        my_device_id = self.getInstanceId()
        bridge_output_choice = BridgeOutputOptions(
            self.unsafe_get("outputForBridgeOutput")
        )

        earliest_source_timestamp = float("inf")
        image_datas, image_masks, module_indices = [], [], []
        for source, (data, source_timestamp) in sources.items():
            earliest_source_timestamp = min(
                earliest_source_timestamp, source_timestamp.toTimestamp()
            )
            # regular TrainMatcher output
            self.output.write(
                data, ChannelMetaData(source, source_timestamp), copyAllData=False
            )
            if bridge_output_choice is BridgeOutputOptions.MATCHED:
                self.zmq_output.write(source, data, source_timestamp)

            if (
                source not in self._source_to_index
                or not data.has(self._image_data_path)
            ):
                continue
            image_data = data[self._image_data_path]
            if isinstance(image_data, ImageData):
                # TODO: maybe glance encoding here
                image_data = image_data.getData()
            # TODO: set dtype based on input?
            image_datas.append(image_data.astype(np.float32, copy=False))
            if data.has(self._image_mask_path):
                image_mask = data[self._image_mask_path]
                if isinstance(image_mask, ImageData):
                    image_mask = image_mask.getData()
            else:
                image_mask = np.zeros_like(image_data, dtype=np.uint8)
            image_masks.append(image_mask)
            module_indices.append(self._source_to_index[source])

        self.output.update(safeNDArray=True)
        if bridge_output_choice is BridgeOutputOptions.MATCHED:
            self.zmq_output.update()

        dims = ["module", "slow_scan", "fast_scan"]
        coords = {"module": module_indices}
        assembled_data, _ = self._geometry.position_modules(
            xr.DataArray(
                data=image_datas,
                dims=dims,
                coords=coords,
            )
        )
        assembled_mask, _ = self._geometry.position_modules(
            xr.DataArray(
                data=image_masks,
                dims=dims,
                coords=coords,
            )
        )
        assembled_mask |= (~np.isfinite(assembled_data))
        output_hash = Hash(
            "image.data",
            assembled_data,
            "image.mask",
            assembled_mask,
        )
        self.assembled_output.write(
            output_hash,
            ChannelMetaData(f"{my_device_id}:assembledOutput", my_timestamp),
            copyAllData=False,
        )
        self.assembled_output.update(safeNDArray=True)
        if bridge_output_choice is BridgeOutputOptions.ASSEMBLED:
            self.zmq_output.write(
                f"{my_device_id}:assembledOutput", output_hash, my_timestamp
            )
            self.zmq_output.update()

        self._preview_friend.write_outputs(
            my_timestamp,
            np.ma.masked_array(data=assembled_data, mask=assembled_mask),
        )

        self._processing_time_tracker.update(default_timer() - ts_start)
        self._tof_tracker.update(Timestamp().toTimestamp() - earliest_source_timestamp)

        self.info["processingTime"] = self._processing_time_tracker.get() * 1000
        self.info["timeOfFlight"] = self._tof_tracker.get() * 1000
        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()

    def preReconfigure(self, conf):
        super().preReconfigure(conf)
        if conf.has("sourceToModuleIndex"):
            selected = [row for row in conf.get("sourceToModuleIndex") if row["select"]]
            assert all(row["index"] >= 0 for row in selected)
            assert len(selected) == len(set(row["index"] for row in selected))
        if any(
            conf.has(key) for key in [
                "sourceToModuleIndex", "sourceIndexPattern", "sources"
            ]
        ):
            self._need_to_update_source_index_mapping = True

        self._preview_friend.reconfigure(conf)

    def postReconfigure(self):
        if self._need_to_update_source_index_mapping:
            self._merge_source_to_index_from_regex()
            self._set_source_to_index_from_table()
