import numpy as np

from karabo.bound import (
    NODE_ELEMENT,
    STRING_ELEMENT,
)

from .base_kernel import BaseArbiterKernel, KernelWarning


class PpuKernel(BaseArbiterKernel):
    _node_name = "ppu"

    def __init__(self, config):
        super().__init__(config)
        self._ppu_device_id = config.get("ppuDevice")
        self._target = 0
        self._num_trains = np.iinfo(np.uint64).max

    @staticmethod
    def extend_device_schema(schema, prefix):
        (
            NODE_ELEMENT(schema)
            .key(prefix)
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{prefix}.ppuDevice")
            .description(
                "Name of the PPU trigger device to follow. It must separately be added "
                "to the 'sources' configuration. Specifically, add "
                "'[ppuDevice].trainTrigger' to the sources such that the relevant "
                "property node is matched."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),
        )

    def consider(self, train_id, sources, num_frames):
        with self.warning_context(KernelWarning.MISSINGSOURCE):
            ppu_data = next(
                data
                for source, (data, _) in sources.items()
                if source == self._ppu_device_id
            )
            with self.warning_context(KernelWarning.MISSINGKEY):
                self._target = ppu_data["trainTrigger.sequenceStart.value"]
                self._num_trains = ppu_data["trainTrigger.numberOfTrains.value"]

        if train_id in range(self._target, self._target + self._num_trains):
            return np.ones(num_frames, dtype=np.uint8)
        else:
            return np.zeros(num_frames, dtype=np.uint8)
