#include <cmath>
#include <nvfunctional>

class MaskedImageFrame {
public:
	const float* data_start;
	const unsigned short num_rows;
	const unsigned short num_cols;
	__device__ MaskedImageFrame(const float* data_start, const unsigned short num_rows, const unsigned short num_cols):
		data_start(data_start), num_rows(num_rows), num_cols(num_cols) {}
	__device__ bool is_masked(const unsigned short i, const unsigned short j) {
		return isnan(data_start[i * num_cols + j]);
	}
	__device__ float get(const unsigned short i, const unsigned short j) {
		return data_start[i * num_cols + j];
	}
	__device__ float get(const unsigned short i, const unsigned short j, const float fallback) {
		if (is_masked(i, j)) {
			return fallback;
		}
		return get(i, j);
	}
	__device__ void maybe_call(const unsigned short i, const unsigned short j, const nvstd::function<void(unsigned short, unsigned short, float)> &fun) {
		if (!is_masked(i, j)) {
			fun(i, j, get(i, j));
		}
	}

	/* Applies fun to all unmasked elements of frame lying on a ring (well, at least
	   a circle in the infinity norm) with given center and radius.
	   TODO: consider making a custom iterator, too
	*/
	__device__ void fun_ring(const unsigned short center_i,
	                         const unsigned short center_j,
	                         const unsigned short radius,
	                         const nvstd::function<void(unsigned short, unsigned short, float)> &fun) {
		if (radius==0) {
			maybe_call(center_i, center_j, fun);
			return;
		}

		// top row
		for (short j=-radius; j<=radius; ++j) {
			maybe_call(center_i - radius, center_j + j, fun);
		}
		// left / right side (not overlapping with top/bottom)
		for (short i=-radius+1; i<radius; ++i) {
			maybe_call(center_i + i, center_j - radius, fun);
			maybe_call(center_i + i, center_j + radius, fun);
		}
		// bottom row
		for (short j=-radius; j<=radius; ++j) {
			maybe_call(center_i + radius, center_j + j, fun);
		}
	}
};

extern "C" __global__ void pf9(const unsigned short num_frames,
                               const unsigned short num_rows,
                               const unsigned short num_cols,
                               const unsigned short window_radius,
                               const float min_peak_over_border,
                               const float min_snr_biggest_pixel,
                               const float min_snr_peak_pixels,
                               const float min_snr_whole_peak,
                               const float min_sigma,
                               const unsigned int max_peaks,
                               const float* image,
                               unsigned int* output_counts,
                               float* output_x,
                               float* output_y,
                               float* output_intensity) {

	// execution model: one thread handles one pixel in one frame
	const unsigned short frame = blockDim.x * blockIdx.x + threadIdx.x;
	const unsigned short row = blockDim.y * blockIdx.y + threadIdx.y;
	const unsigned short col = blockDim.z * blockIdx.z + threadIdx.z;

	if (frame >= num_frames ||
	    row < window_radius || row >= num_rows - window_radius ||
	    col < window_radius || col >= num_cols - window_radius) {
		return;
	}

	// wrap thin helper class around for convenience
	MaskedImageFrame masked_frame(image + frame * (num_rows * num_cols),
	                              num_rows,
	                              num_cols);

	// candidate should not be masked
	if (masked_frame.is_masked(row, col)) {
		return;
	}

	float pixel_val = masked_frame.get(row, col);

	// candidate should be greater than immediate neighbors
	// with tie breaking: in case of equality, lowest row, col is candidate
	if (pixel_val <= masked_frame.get(row-1, col-1, -INFINITY) ||
	    pixel_val <= masked_frame.get(row-1, col, -INFINITY) ||
	    pixel_val <= masked_frame.get(row-1, col+1, -INFINITY) ||
	    pixel_val <= masked_frame.get(row, col-1, -INFINITY) ||
	    pixel_val < masked_frame.get(row, col+1, -INFINITY) ||
	    pixel_val < masked_frame.get(row+1, col, -INFINITY) ||
	    pixel_val < masked_frame.get(row+1, col, -INFINITY) ||
	    pixel_val < masked_frame.get(row+1, col, -INFINITY)) {
		return;
	}

	// candidate should be greater than pixels on border
	// note: original PF9 only checks three pixels per side around window
	// (full border for window radius 2, less for higher)
	{
		float border_max = -INFINITY;
		masked_frame.fun_ring(row, col, window_radius, [&] (unsigned short, unsigned short, float val) {
			border_max = max(border_max, val);
		});
		if (pixel_val - min_peak_over_border <= border_max) {
			return;
		}
	}

	// candidate should have sufficient SNR over border
	float mean = 0;
	float sigma = 0;
	{
		// https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm
		unsigned short count = 0;
		float M2 = 0;
		masked_frame.fun_ring(row, col, window_radius, [&] (unsigned short, unsigned short, float val) {
			++count;
			float delta = val - mean;
			mean += delta / static_cast<float>(count);
			float delta2 = val - mean;
			M2 += delta * (delta2);
		});
		if (count < 10) {
			// should this be configurable? (check for validPixelCount < 10 is hardcoded in PF9)
			return;
		}
		// TODO: form opinion on /(n-1) for dealing with biased sample stdev
		sigma = max(min_sigma, sqrtf(M2 / (static_cast<float>(count) - 1)));

		if (pixel_val <= mean + min_snr_biggest_pixel * sigma) {
			return;
		}
	}

	// whole peak should have sufficent SNR
	float peak_weighted_row;
	float peak_weighted_col;
	float peak_total_mass = pixel_val;
	{
		/* TODO: more compact form */
		float peak_weighted_row_nom = static_cast<float>(row) * pixel_val;
		float peak_weighted_col_nom = static_cast<float>(col) * pixel_val;
		const float peak_pixel_threshold = mean + min_snr_peak_pixels * sigma;
		for (unsigned short layer=1; layer<=window_radius; ++layer) {
			float total_mass_before = peak_total_mass;
			masked_frame.fun_ring(row, col, layer, [&] (unsigned short i, unsigned short j, float val) {
				if (val > peak_pixel_threshold) {
					peak_total_mass += val;
					peak_weighted_row_nom += val * static_cast<float>(i);
					peak_weighted_col_nom += val * static_cast<float>(j);
				}
			});
			// in case nothing was added, stop expanding
			if (peak_total_mass == total_mass_before) {
				break;
			}
		}
		if (peak_total_mass <= mean + min_snr_whole_peak * sigma) {
			return;
		}
		peak_weighted_row = peak_weighted_row_nom / peak_total_mass;
		peak_weighted_col = peak_weighted_col_nom / peak_total_mass;
	}

	unsigned int output_index = atomicInc(output_counts + frame, max_peaks);
	unsigned int output_pos = frame * max_peaks + output_index;
	output_x[output_pos] = peak_weighted_row;
	output_y[output_pos] = peak_weighted_col;
	output_intensity[output_pos] = peak_total_mass;
}
