import numpy as np

from calng import base_kernel_runner, utils


pixel_data = np.arange(512 * 1024).reshape(512, 1024).astype(np.float32)


def test_cols_A0123():
    def convert_cols_A0123(data, out=None):
        # https://redmine.xfel.eu/issues/126444
        if out is None:
            out = np.zeros((86, (1024 * 3 + 18)), dtype=np.float32)
        # 256 not divisible by 3, so we round up
        # 18 since we have 6 more pixels in H per gap
        # first we fill the normal pixels, the gap ones will be overwritten later
        for yin in range(256):
            for xin in range(1024):
                ichip = int(xin / 256)
                xout = (ichip * 774) + (xin % 256) * 3 + yin % 3
                # 774 is the chip period, 256*3+6
                yout = int(yin / 3)
                out[yout, xout] = data[yin, xin]
        # now the gap pixels...
        for igap in range(3):
            for yin in range(256):
                yout = int(yin / 6) * 2
                # first the left side of gap
                xin = igap * 256 + 255
                xout = igap * 774 + 765 + yin % 6
                out[yout, xout] = data[yin, xin]
                out[yout + 1, xout] = data[yin, xin]
                # then the right side is mirrored
                xin = igap * 256 + 255 + 1
                xout = igap * 774 + 765 + 11 - yin % 6
                out[yout, xout] = data[yin, xin]
                out[yout + 1, xout] = data[yin, xin]
            # out[yout, xout] = out[yout, xout] / 2
            # if we want a proper normalization
            # (the area of those pixels is double, so they see 2x the signal)
        return out

    strixel_naive = convert_cols_A0123(pixel_data)
    lut_package = np.load(
        base_kernel_runner.kernel_dir / "strixel_cols_A0123-lut_mask.npz"
    )
    assert tuple(lut_package["frame_shape"]) == strixel_naive.shape
    strixel_lut = np.zeros_like(strixel_naive)
    utils.apply_partial_lut(
        data=pixel_data,
        lut=lut_package["lut"],
        mask=lut_package["mask"],
        out=strixel_lut,
        missing=0,
    )
    assert np.array_equal(strixel_naive, strixel_lut)


def test_rows_A1256():
    def convert_rows_A1256(data, out=None):
        # https://redmine.xfel.eu/issues/161148
        if out is None:
            out = np.zeros([*data.shape[:-2], 1488, 165], dtype=np.float32)

        # Select only 4 center ASICS
        data = data[..., 256:768]

        # Offset due to guard ring pixels
        offset_y = 9
        offset_x = 11

        for xin in range(
            offset_x, 512 - 9
        ):  # on the right side, there are 9 guard ring pixels
            # Bottom ASICs
            for yin in range(offset_y, 255):
                yout = (xin - offset_x) % 3 + (yin - offset_y) * 3
                xout = (xin - offset_x) // 3 + (xin) // 257
                out[..., yout, xout] = data[..., yin, xin]

            # Top ASICs (mirrored on the horizontal axis)
            for yin in range(257, 512 - offset_y):
                yout = 2 - (xin - offset_x) % 3 + (yin - offset_y) * 3 + 6
                xout = (xin - offset_x) // 3 + (xin) // 257
                out[..., yout, xout] = data[..., yin, xin]

        return out

    strixel_naive = convert_rows_A1256(pixel_data)
    lut_package = np.load(
        base_kernel_runner.kernel_dir / "strixel_rows_A1256-lut_mask.npz"
    )
    assert tuple(lut_package["frame_shape"]) == strixel_naive.shape
    strixel_lut = np.zeros_like(strixel_naive)
    utils.apply_partial_lut(
        data=pixel_data,
        lut=lut_package["lut"],
        mask=lut_package["mask"],
        out=strixel_lut,
        missing=0,
    )
    assert np.array_equal(strixel_naive, strixel_lut)
